// Firmware for "C118MB"-type smart sockets.
// (C) 2019 by miq1@gmx.de
//
// Made based on the findings of PeteKnight as seen on https://community.blynk.cc/t/alternative-to-sonoff-s20-eu-type-f-smart-socket/23318
//
// Features:
// * Imitates three Belkin WeMo Smart switches (courtesy of the fauxmoESP library: https://bitbucket.org/xoseperez/fauxmoesp/downloads/)
// * ON/OFF by pressing the button on the socket once (switch 1), twice (switch 2) or thrice/long (both switches)
// * ON states signalled by steady blue(switch 1) or green (switch 2) LEDs
// * OTA-Update from Arduino IDE
// * Config mode - started automatically on freshly setup devices or by pressing the button within 3s from system reset
//   WIFI access point SSID: "Socket_XXXXXX" with XXXXXX=ESP flash ID, password "Maelstrom"
//   Web page on http://192.168.4.1 for configuration
//   Configurable values:
//   * SSID for home network
//   * Password for home network
//   * Device IDs for both WeMo and OTA names
//   * OTA password
//   Config data permanently stored in SPIFFS

// Library includes
#include <Arduino.h>
#include <ArduinoOTA.h>
#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include "fauxmoESP.h"
#include <Ticker.h>
#include <mdButton.h>
#include "FS.h"

// GPIO definitions
#define LED 13
#define BUTTON1_LED 5
#define RELAY1 12
#define BUTTON2_LED 16
#define RELAY2 4
#define BUTTON_PIN 14
#define LED_OFF HIGH
#define LED_ON LOW
#define RELAY_OFF LOW
#define RELAY_ON HIGH

mdButton button(BUTTON_PIN);

// Operations modes
#define RUN 1
#define CONFIG 2

// maximum length of configuration parameters
#define PARMLEN 64

// Blink patterns for the different modes
// CONFIG mode:
#define CONFIGBLINK 0xAAA0
// WIFI connect phase:
#define WIFIBLINK 0x0303
// SPIFFS formatting phase:
#define INITBLINK 0x000F

// File paths in SPIFFS
#define SSIDFILE "/conf/SSID"
#define PWDFILE  "/conf/PWD"
#define DEV1FILE  "/conf/DEV1"
#define DEV2FILE  "/conf/DEV2"
#define DEV3FILE  "/conf/DEV3"
#define OHOSTFILE "/conf/OHOST"
#define OPWDFILE "/conf/OTA"

fauxmoESP fauxmo;             // create WeMo lookalike
Ticker redticker;            // Ticker maintaining the red LED blinking
ESP8266WebServer server(80);  // Web server on port 80
uint8_t mode;                 // Operations mode, RUN or CONFIG

// char arrays for configuration parameters
char C_SSID[PARMLEN];
char C_PWD[PARMLEN];
char O_HOST[PARMLEN];
char O_PWD[PARMLEN];

// data structure for three Wemo devices (left, right and both switches)
struct {
  char name[PARMLEN];  // Device name
  uint8_t id;          // Device ID
  bool state;          // Status on/off
  uint8_t led;         // led port
  uint8_t relay;       // relay port
} device[3] = {
  {"LEFT", 1, false, BUTTON1_LED, RELAY1 },
  {"RIGHT", 2, false, BUTTON2_LED, RELAY2 },
  {"BOTH", 3, false, BUTTON1_LED, RELAY1 }
};

// -----------------------------------------------------------------------------
// Setup WiFi in RUN mode
// -----------------------------------------------------------------------------
void wifiSetup() {
    // Set WIFI module to STA mode
    WiFi.mode(WIFI_STA);

    // Connect
    WiFi.begin(C_SSID, C_PWD);

    // Wait
    while (WiFi.status() != WL_CONNECTED) {
        delay(100);
    }
    // Connected!
}

// -----------------------------------------------------------------------------
// Change state of WeMo device ON<-->OFF
// -----------------------------------------------------------------------------
void SetState(uint8_t device_id, const char * device_name, bool state) 
{
  for(uint8_t i=0; i<3; ++i)
  {
    if(device[i].id == device_id)
    {
      // found it!
      if(i==2) // both switches
      {
        digitalWrite(device[0].led, state ? LED_ON : LED_OFF);
        digitalWrite(device[0].relay, state ? RELAY_ON : RELAY_OFF);
        device[0].state = state;
        digitalWrite(device[1].led, state ? LED_ON : LED_OFF);
        digitalWrite(device[1].relay, state ? RELAY_ON : RELAY_OFF);
        device[1].state = state;
      }
      else
      {
        digitalWrite(device[i].led, state ? LED_ON : LED_OFF);
        digitalWrite(device[i].relay, state ? RELAY_ON : RELAY_OFF);
      }
      device[i].state = state; 
      break;
    }
  }
}

// -----------------------------------------------------------------------------
// Report state when asked
// -----------------------------------------------------------------------------
bool GetState(uint8_t device_id, const char * device_name) 
{
  for(uint8_t i=0; i<3; ++i)
  {
    if(device[i].id == device_id)
    {
      // found it!
      return device[i].state;
    }
  }
}

// -----------------------------------------------------------------------------
// Setup. Find out which mode to run and initialize objects
// -----------------------------------------------------------------------------
void setup() 
{
  File f;

  // Define GPIO input/output direction
  pinMode(LED, OUTPUT);
  pinMode(RELAY1,OUTPUT);
  pinMode(BUTTON1_LED,OUTPUT);
  pinMode(RELAY2,OUTPUT);
  pinMode(BUTTON2_LED,OUTPUT);

  // initially switch to OFF
  digitalWrite(RELAY1, RELAY_OFF); // Relay OFF
  digitalWrite(RELAY2, RELAY_OFF); // Relay OFF
  digitalWrite(BUTTON1_LED, LED_OFF);  // LED OFF
  digitalWrite(BUTTON2_LED, LED_OFF);  // LED OFF

  // Open SPIFFS.
  bool rc = SPIFFS.begin();

  // Open failed or no formating registered?
  if(!rc || !SPIFFS.exists("/fmtOK.txt"))
  {
    // YES. Start special blink pattern and format SPIFFS
    redticker.attach(0.1, redtick, INITBLINK);
    SPIFFS.format();

    // Now register format to do it only once
    f = SPIFFS.open("/fmtOK.txt", "w");
    f.print("Format OK");
    f.close();

    // stop blinking
    redticker.detach();
  }

  // attempt to read all configuration values. Count each successful read.
  uint8_t confcnt = 0;
  if(getOut(C_SSID, SSIDFILE)) confcnt++;
  if(getOut(C_PWD, PWDFILE)) confcnt++;
  if(getOut(device[0].name, DEV1FILE)) confcnt++;
  if(getOut(device[1].name, DEV2FILE)) confcnt++;
  if(getOut(device[2].name, DEV3FILE)) confcnt++;
  if(getOut(O_HOST, OHOSTFILE)) confcnt++;
  if(getOut(O_PWD, OPWDFILE)) confcnt++;

  // Determine operation mode - normally RUN
  mode = RUN;

  // we will wait 3s for button presses to deliberately enter CONFIG mode
  long t0 = millis()+3000;

  while(millis()<t0)
  {
    // Button pressed?
    if(button.status()>0)
    {
      // YES. Force CONFIG mode and leave. 
      confcnt = 0;
      break;
    }
  }

  // CONFIG mode required?
  if(confcnt<7)
  {
    // YES. Switch to CONFIG and start special blink pattern,
    mode = CONFIG;
    redticker.attach(0.1, redtick, CONFIGBLINK);

    // Set up access point mode. First, create AP SSID from flash ID,
    char APssid[64];
    strcpy(APssid, "Socket_XXXXXX");
    long id = ESP.getChipId();
    for(uint8_t i=6; i; i--)
    {
      char c = (id & 0xf);
      if(c>9) { c -= 10; c += 'A'; }
      else    c += '0';
      APssid[6+i] = c;
      id >>= 4;
    }

    // Start AP.
    WiFi.softAP(APssid, "Maelstrom");

    // Register URLs for functions on web page.
    server.on("/", handleRoot);         // Main page
    server.on("/reset", handleRestart); // Reset button on page
    server.on("/save", handleSave);     // Save button on page

    // Start web server.
    server.begin();
  }
  else
  {
    // NO, RUN mode.
    // Set blink pattern for WiFi connection phase.
    redticker.attach(0.1, redtick, WIFIBLINK);

    // Connect. Will not terminate if not successful!!!
    wifiSetup();

    // Stop blinking.
    redticker.detach();
    
    // Fauxmo setup.
    device[0].id = fauxmo.addDevice(device[0].name);
    digitalWrite(device[0].led, LED_OFF);
    digitalWrite(device[0].relay, RELAY_OFF);
    device[0].state = false;
    device[1].id = fauxmo.addDevice(device[1].name);
    digitalWrite(device[1].led, LED_OFF);
    digitalWrite(device[1].relay, RELAY_OFF);
    device[1].state = false;
    device[2].id = fauxmo.addDevice(device[2].name);
    device[2].state = false;
    
    fauxmo.enable(true);              // get visible.
    fauxmo.onSetState(SetState);      // link to switch callback
    fauxmo.onGetState(GetState);      // link to inquire callback

    // ArduinoOTA setup
    ArduinoOTA.setHostname((const char *)O_HOST);  // Set OTA host name
    ArduinoOTA.setPassword((const char *)O_PWD);  // Set OTA password
    ArduinoOTA.begin();               // start OTA scan
  }
}

// -----------------------------------------------------------------------------
// Main loop
// -----------------------------------------------------------------------------
void loop() 
{
  // RUN mode?
  if(mode==RUN)
  {
    // YES. Check WeMo, OTA and button requests.
    fauxmo.handle();
    ArduinoOTA.handle();

    // If button pressed, toggle Relay/LED
    int n = button.status();
    if(n==1)  // single press: left socket
    {
      SetState(device[0].id, device[0].name, !device[0].state);
    }
    else if(n==2) // double press: right socket
    {
      SetState(device[1].id, device[1].name, !device[1].state);
    }
    else if(n==3) // triple press: both sockets ON
    {
      SetState(device[0].id, device[0].name, true);
      SetState(device[1].id, device[1].name, true);
    }
    else if(n<0) // long press: both sockets OFF
    {
      SetState(device[0].id, device[0].name, false);
      SetState(device[1].id, device[1].name, false);
    }
  }
  
  else
  {
    // NO. CONFIG mode - listen to web requests only.
    server.handleClient();
  }
}

// -----------------------------------------------------------------------------
// redtick - maintain blink patterns
// -----------------------------------------------------------------------------
void redtick(int pattern)
{
  static uint8_t counter = 0;       // Counter to check 16 pattern bits in sequence
  
  // Toggle state, if necessary
  int state = digitalRead(LED);       // get the current state of the LED pin
  int bite = ((1<<counter)&pattern);  // get the current bit of the blink pattern
  if(bite) // if HIGH - 
  {
    if(state) digitalWrite(LED, LED_ON); // switch LED ON, if was OFF
  }
  else // if LOW - 
  {
    if(!state) digitalWrite(LED, LED_OFF); // switch LED OFF, if was ON
  }
  counter++;                          // advance one bit
  counter &= 0xf;                     // catch overflow - counter runs 0-15
}

// -----------------------------------------------------------------------------
// handleRoot - bring out configuration page
// -----------------------------------------------------------------------------
void handleRoot()
{
  // Setup HTML with embedded values from SPIFFS - if any.
  String page = "<!DOCTYPE html><html lang=\"en\"><head><meta charset=\"utf-8\">" \
    "<title>Smart socket setup</title>" \
    "<style type=\"text/css\">" \
    "label { display: block; width: 200px; font-size: small; }" \
    "legend { font-weight: bold; }" \
    "input.button { width: 10em;  height: 3em; font-weight: bold; }" \
    "table {border: none;}" \
    "</style></head><body><h1>Socket setup</h1>" \
    "<form><fieldset style=\"background-color:#FFEFD5\"><legend>WiFi network</legend>" \
    "<label for=\"ssid\">SSID</label><input type=\"text\" id=\"ssid\" name=\"ssid\" maxlength=32 size=40 required value=\"" 
    + String(C_SSID) + // merge in home network SSID
    "\"><br/>" \
    "<label for=\"pwd\">Password</label><input type=\"password\" id=\"pwd\" name=\"pwd\" maxlength=32 size=40 value=\""
    + String(C_PWD) + // merge in home network password
    "\">" \
    "</fieldset><p/><fieldset style=\"background-color:#DCDCDC\"><legend>Device settings</legend>" \
    "<label for=\"device1\">Left device name</label><input type=\"text\" id=\"device1\" name=\"device1\" maxlength=32 size=40 pattern=\"[A-Za-z0-9_-]+\" required value=\""
    + String(device[0].name) +  // merge in WeMo/OTA device name
    "\"><br/>" \
    "<label for=\"device2\">Right device name</label><input type=\"text\" id=\"device2\" name=\"device2\" maxlength=32 size=40 pattern=\"[A-Za-z0-9_-]+\" required value=\""
    + String(device[1].name) +  // merge in WeMo/OTA device name
    "\"><br/>" \
    "<label for=\"device3\">Both devices name</label><input type=\"text\" id=\"device3\" name=\"device3\" maxlength=32 size=40 pattern=\"[A-Za-z0-9_-]+\" required value=\""
    + String(device[2].name) +  // merge in WeMo/OTA device name
    "\"><br/>" \
    "<label for=\"otahost\">OTA Hostname</label><input type=\"text\" id=\"otahost\" name=\"otahost\" maxlength=32 size=40 value=\""
    + String(O_HOST) +  // merge in OTA hostname
    "\"><br/>" \
    "<label for=\"otapwd\">OTA Password</label><input type=\"text\" id=\"otapwd\" name=\"otapwd\" maxlength=32 size=40 value=\""
    + String(O_PWD) +  // merge in OTA password
    "\">" \
    "</fieldset><p/>" \
    "<input type=\"submit\" value=\"Save\" name=\"send\" formaction=\"/save\" class=\"button\" style=\"color:black;background-color:#32CD32\">" \
    "</form><p/><table><tr><td>ESP ID</td><td>"
    + String(ESP.getFlashChipId(), HEX) + 
    "</td></tr><tr><td>Speed</td><td>" 
    + String(ESP.getFlashChipSpeed()) + 
    "</td></tr><tr><td>Flash size</td><td>" 
    + String(ESP.getFlashChipRealSize()) + 
    "</td></tr><tr><td>Flash mode</td><td>" 
    + String(ESP.getFlashChipMode()) + 
    "</td></tr></table><p/><form>" \
    "<input type=\"submit\" value=\"Reset\" name=\"send\" formaction=\"/reset\" class=\"button\" style=\"color:white;background-color:#FF4500\">" \
    "</form></body></html>";

  // Send out page
  server.send(200, "text/html", page);
}

// -----------------------------------------------------------------------------
// handleRestart - reset device on web page request (user pressed Reset button)
// -----------------------------------------------------------------------------
void handleRestart()
{
  ESP.restart();
}

// -----------------------------------------------------------------------------
// handleSave - get user input from web page and store in SPIFFS and config parameters
// -----------------------------------------------------------------------------
void handleSave()
{
  fillIn(C_SSID, server.arg("ssid"), SSIDFILE);
  fillIn(C_PWD, server.arg("pwd"), PWDFILE);
  fillIn(device[0].name, server.arg("device1"), DEV1FILE);
  fillIn(device[1].name, server.arg("device2"), DEV2FILE);
  fillIn(device[2].name, server.arg("device3"), DEV3FILE);
  fillIn(O_HOST, server.arg("otahost"), OHOSTFILE);
  fillIn(O_PWD, server.arg("otapwd"), OPWDFILE);

  // re-display configuration web page
  handleRoot();
}

// -----------------------------------------------------------------------------
// fillIn - saves String arg into SPIFFS and configuration parameter
// -----------------------------------------------------------------------------
bool fillIn(char *target, String arg, char *writename)
{
  File f;

  // No value? Bail out!
  if(!arg) return false;

  // get length of given parameter
  uint8_t len = arg.length();

  // Copy to parameter char array
  arg.toCharArray(target, PARMLEN);

  // Also write into SPIFFS 
  f = SPIFFS.open(writename, "w");
  if(f)
  {
    f.write((uint8_t *)target, len);
    f.close();
  }
  return true;
}

// -----------------------------------------------------------------------------
// getOut - read configuration parameter from SPIFFS and fill in char array
// -----------------------------------------------------------------------------
bool getOut(char *target, char *readname)
{
  File f;
  
  *target = 0;                    // assume no data
  
  // Open SPIFFS path
  f = SPIFFS.open(readname, "r");
  if(f)  // valid file?
  {
    // YES. get data length and read bytes into char array.
    uint8_t len = f.size();
    f.read((uint8_t *)target, len);
    f.close();

    // terminate char array with 0x00
    target[len] = 0;

    // leave OK
    return true;
  }

  // NO, no file found. Bail out!
  return false;
}
